@@ -565,6 +565,22 @@ gil_unselect_all (GnomeIconList *gil, GdkEvent *event, gpointer keep)
 	return idx;
 }
 
+/**
+ * gnome_icon_list_select_all:
+ * @gil:   An icon list.
+ */
+void
+gnome_icon_list_select_all (GnomeIconList *gil)
+{
+	uint i;
+	
+	g_return_if_fail (gil != NULL);
+	g_return_if_fail (IS_GIL (gil));
+	
+	for (i = 0; i < gnome_icon_list_get_num_icons (gil); i++)
+		gnome_icon_list_select_icon (gil, i);
+}
+
 /**
  * gnome_icon_list_unselect_all:
  * @gil:   An icon list.
@@ -884,7 +900,7 @@ icon_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
 
 	if (priv->editing_icon && event->type == GDK_BUTTON_PRESS) {
 		gnome_icon_text_item_stop_editing (priv->editing_icon->text, FALSE);
-		priv->editing_icon = FALSE;
+		priv->editing_icon = NULL;
 	}
 
 	switch (priv->selection_mode) {
@@ -1645,6 +1661,19 @@ real_move_cursor (Gil *gil, GtkDirectionType dir, gboolean clear_selection)
 		if (priv->focus_icon - items_per_line >= 0)
 			new_focus_icon -= items_per_line;
 		break;
+
+	case GTK_SCROLL_START:
+		if (priv->focus_icon != 0) {
+			new_focus_icon = 0;
+		}
+		break;
+
+	case GTK_SCROLL_END:
+		if (priv->focus_icon != (priv->icons -1)) {
+			new_focus_icon = priv->icons -1;
+		}
+		break;
+
 	default:
 		break;
 	}
@@ -1656,7 +1685,7 @@ real_move_cursor (Gil *gil, GtkDirectionType dir, gboolean clear_selection)
 	}
 
 	if (gnome_icon_list_icon_is_visible (gil, new_focus_icon) != GTK_VISIBILITY_FULL &&
-	    (dir == GTK_DIR_UP || dir == GTK_DIR_DOWN))
+	    (dir == GTK_DIR_UP || dir == GTK_DIR_DOWN || GTK_SCROLL_START || GTK_SCROLL_END))
 		gnome_icon_list_moveto (gil, new_focus_icon, dir == GTK_DIR_UP ? 0.0 : 1.0);
 
 	gnome_icon_list_focus_icon (gil, new_focus_icon);
@@ -2317,6 +2346,11 @@ gnome_icon_list_class_init (GilClass *gil_class)
 	add_move_binding (binding_set, GDK_Down, GTK_DIR_DOWN);
 	add_move_binding (binding_set, GDK_Up, GTK_DIR_UP);
 
+	add_move_binding (binding_set, GDK_Home, GTK_SCROLL_START);
+	add_move_binding (binding_set, GDK_KP_Home, GTK_SCROLL_START);
+	add_move_binding (binding_set, GDK_End, GTK_SCROLL_END);
+	add_move_binding (binding_set, GDK_KP_End, GTK_SCROLL_END);
+
 	gtk_binding_entry_add_signal (binding_set, GDK_space, 0, "toggle_cursor_selection", 0);
 	gtk_binding_entry_add_signal (binding_set, GDK_space, GDK_CONTROL_MASK, "toggle_cursor_selection", 0);
 
